import numpy as np
from pyTurbulence.syntheticTurbulence import *
from pyTurbulence.spectrum import compute_tke_spectrum, plot_spectrum, energy_spectrum
from pyTurbulence.plot3D import plot3D
from datetime import datetime
import time
import os

def solve(user_params)->dict:
    """
    Solve the synthetic turbulence problem.

    Parameters
    ----------
    params : dict
        Dictionary containing the simulation parameters. Expected keys and descriptions:
        
        - "nbModes" (int): Number of modes per wave (default: 128).
        - "Mt" (float): Turbulent Mach number (default: 0.2).
        - "Pressure" (float): Mean pressure in the simulation (default: 1.0).
        - "Temperature" (float): Mean temperature in the simulation (default: 1.0).
        - "k0" (int): Peak wavenumber (default: 4).
        - "domain_size" (list, tuple, np.ndarray): Size of the simulation domain, should be a 3D vector (default: [2π, 2π, 2π]).
        - "domain_resolution" (list, tuple, np.ndarray): Resolution of the domain, should be a 3D vector (default: [64, 64, 64]).
        - "Spectrum" (str): Type of spectrum for turbulence generation. Options: "PassotPouquet" (default: "PassotPouquet").
        - "gamma" (float): Heat capacity ratio, typically between 1 and 2 (default: 1.4).
        - "case" (int): Type of turbulent fields to generate. Options: 0, 1 (default: 1). 0: incompressible - 1: compressible.
        - "output_dir" (str): Directory to store output results (default: "results").

    Returns
    -------
    results : dict
        Dictionary containing the results of the simulation.
    """    

    # Validate the input parameters
    params = _validate_params(user_params)

    # Read the params dictionary
    nbModes = params["nbModes"]
    Mt = params["Mt"]
    mean_pressure = params["Pressure"]
    mean_temperature = params["Temperature"]
    k0 = params["k0"]
    domain_size = params["domain_size"]
    domain_resolution = params["domain_resolution"]
    spectrum = params["Spectrum"]
    gamma = params["gamma"]
    results_dir = params["output_dir"]

    # Create results directory
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)

    # Compute additional parameters
    mean_density = mean_pressure / mean_temperature
    celerity = np.sqrt(gamma * mean_pressure / mean_density)
    Ek = 0.5 * (Mt * celerity) ** 2
    urms = np.sqrt(2. * Ek / 3.)

    # Generate the solenoidal velocity field
    start = time.time()
    us, vs, ws, wmax = compute_solenoidal_velocities(spectrum, nbModes, urms, 
                                                  k0, domain_size, 
                                                  domain_resolution)
    TKE = np.mean(0.5 * (us.reshape(-1) ** 2 + vs.reshape(-1) ** 2 + ws.reshape(-1) ** 2))
    end = time.time()
    solenoidal_time = end - start

    # Generate the incompressible pressure fluctuations
    start = time.time()
    incompressible_pressure_fluctuations = compute_solenoidal_pressure(us, vs, ws, 
                                                                       domain_resolution, 
                                                                       domain_size)
    end = time.time()
    pressure_time = end - start

    # Generate the dilatational velocity field
    start = time.time()
    ud, vd, wd = compute_dilatational_velocities(us, vs, ws, 
                                                 incompressible_pressure_fluctuations, 
                                                 gamma, domain_resolution, domain_size)
    end = time.time()
    dilatational_time = end - start

    # Compute the thermodynamic fields
    start = time.time()
    density, pressure, temperature, u, v, w = compute_thermodynamic_fields(mean_density, 
                                                                           mean_pressure, 
                                                                           mean_temperature, 
                                                                           us,vs,ws,
                                                                           ud,vd,wd,
                                                                           incompressible_pressure_fluctuations, 
                                                                           gamma, Mt, params["case"])
    end = time.time()
    thermodynamic_time = end - start

    # Compute the spectrum
    start = time.time()
    knyquist, wave_numbers, tke_spectrum_solenoidal = compute_tke_spectrum(us, vs, ws, 
                                                                            domain_size[0], 
                                                                            domain_size[1], domain_size[2])

    _, _,                 tke_spectrum_dilatational = compute_tke_spectrum(ud, vd, wd, 
                                                                            domain_size[0], 
                                                                            domain_size[1], domain_size[2])
    real_spectrum = energy_spectrum(spectrum, wave_numbers, urms, k0)
    solenoidal_spectrum_path    = os.path.join(results_dir, "spectrum_solenoidal.pdf")
    dilatational_spectrum_path  = os.path.join(results_dir, "spectrum_dilatational.pdf")
    plot_spectrum(wmax, wave_numbers, tke_spectrum_solenoidal,   real_spectrum=real_spectrum, save_path=solenoidal_spectrum_path)
    plot_spectrum(wmax, wave_numbers, tke_spectrum_dilatational, real_spectrum=real_spectrum, save_path=dilatational_spectrum_path)
    end = time.time()
    spectrum_time = end - start

    # Save the spectrum data
    spectrum_data_path = os.path.join(results_dir, "spectrum.dat")
    np.savetxt(spectrum_data_path, np.column_stack((wave_numbers, tke_spectrum_solenoidal)), header="wave_numbers tke_spectrum")

    # Save the data
    start = time.time()
    u.tofile(os.path.join(results_dir, "u.dat"))
    v.tofile(os.path.join(results_dir, "v.dat"))
    w.tofile(os.path.join(results_dir, "w.dat"))
    pressure.tofile(os.path.join(results_dir, "p.dat"))
    temperature.tofile(os.path.join(results_dir, "T.dat"))
    end = time.time()
    save_time = end - start

    results = {
        "TKE": TKE,
        "Ek": Ek,
        "solenoidal_time": solenoidal_time,
        "pressure_time": pressure_time,
        "dilatational_time": dilatational_time,
        "thermodynamic_time": thermodynamic_time,
        "spectrum_time": spectrum_time,
        "save_time": save_time
    }

    # Write the log file
    _write_log_file(params, results)

    # Dictionary containing the results of the simulation
    data = {
        "u": u,
        "v": v,
        "w": w,
        "pressure": pressure,
        "temperature": temperature,
        "wave_numbers": wave_numbers,
        "tke_spectrum_solenoidal": tke_spectrum_solenoidal,
        "tke_spectrum_dilatational": tke_spectrum_dilatational,
        "real_spectrum": real_spectrum
    }


    return data

def _validate_params(params) -> dict:
    """
    Validate the input parameters dictionary, set default values where needed, enforce allowed values, and check numeric ranges.

    Parameters
    ----------
    params : dict
        Dictionary containing the simulation parameters.

    Returns
    -------
    dict
        Validated parameters with defaults set where necessary.

    Raises
    ------
    TypeError
        If a parameter has an incorrect type.
    ValueError
        If a list parameter does not have the correct length or an invalid value.
    """

    # Define required parameters with default values
    default_params = {
        "nbModes": 128,  
        "Mt": 0.2,  
        "Pressure": 1.0,  
        "Temperature": 1.0,  
        "k0": 4,  
        "domain_size": [2.0 * np.pi, 2.0 * np.pi, 2.0 * np.pi],  
        "domain_resolution": [64, 64, 64],  
        "Spectrum": "PassotPouquet",  
        "gamma": 1.4,  
        "case": 1,
        "output_dir": "results"
    }

    # Define expected types for validation
    expected_types = {
        "nbModes": int,
        "Mt": (int, float),
        "Pressure": (int, float),
        "Temperature": (int, float),
        "k0": (int, float),
        "domain_size": (list, tuple, np.ndarray),
        "domain_resolution": (list, tuple, np.ndarray),
        "Spectrum": str,
        "gamma": (int, float),
        "case": int,
        "output_dir": str
    }

    # Define allowed values for specific parameters
    allowed_values = {
        "Spectrum": {"PassotPouquet"},  # Allowed spectrum types
        "case": {0, 1},  # Allowed integer cases
    }

    # Define numeric ranges for specific parameters
    numeric_ranges = {
        "nbModes": (1, None),  # Must be >= 1
        "Mt": (0, None),  # Typical turbulent Mach number is between 0 and 1
        "Pressure": (0, None),  # Pressure must be positive
        "Temperature": (0, None),  # Temperature must be positive
        "k0": (1, None),  # Characteristic wavenumber must be >= 1
        "gamma": (0, None),  # Heat capacity ratio is typically between 1 and 2
    }

    validated_params = default_params.copy()

    for param, expected_type in expected_types.items():
        if param not in params:
            print(f"⚠️ Warning: '{param:<20}' is missing. Using default value: {default_params[param]}")
        else:
            validated_params[param] = params[param]  # Overwrite with user-provided value

            # Type validation
            if not isinstance(validated_params[param], expected_type):
                raise TypeError(
                    f"Error: Incorrect type for '{param}'. "
                    f"Expected {expected_type}, got {type(validated_params[param])}."
                )

            # Allowed values validation
            if param in allowed_values and validated_params[param] not in allowed_values[param]:
                raise ValueError(
                    f"Error: Invalid value for '{param}'. Allowed values: {allowed_values[param]}. "
                    f"Got '{validated_params[param]}'."
                )

            # Numeric range validation
            if param in numeric_ranges:
                min_val, max_val = numeric_ranges[param]
                if min_val is not None and validated_params[param] < min_val:
                    raise ValueError(
                        f"Error: '{param}' must be >= {min_val}. Got {validated_params[param]}."
                    )
                if max_val is not None and validated_params[param] > max_val:
                    raise ValueError(
                        f"Error: '{param}' must be <= {max_val}. Got {validated_params[param]}."
                    )

    # Additional checks for array-like parameters
    if len(validated_params["domain_size"]) != 3:
        raise ValueError("Error: 'domain_size' must be a list or tuple of length 3.")

    if len(validated_params["domain_resolution"]) != 3:
        raise ValueError("Error: 'domain_resolution' must be a list or tuple of length 3.")

    return validated_params

def _get_metadata()->dict:
    metadata = {
        "code_name": "~~~ pyTurbulence: Synthetic Turbulence Generator ~~~",
        "author": "Amaury Bilocq",
        "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    }
    return metadata

def _write_log_file(params, results):
    """
    Write the log file with metadata, parameters, and results.

    Parameters
    ----------
    log_file : str
        Path to the log file.
    params : dict
        Dictionary containing the simulation parameters.
    results : dict
        Dictionary containing the results of the simulation.
    """
    log_file = os.path.join(params["output_dir"], "log.txt")
    metadata = _get_metadata()

    with open(log_file, "w") as f:
        # Display metadata
        f.write(f"{'='*70}\n")
        f.write(f"{metadata['code_name']:^70}\n")
        f.write(f"{'='*70}\n")
        f.write(f"{'Author:':<20} {metadata['author']}\n")
        f.write(f"{'Date:':<20} {metadata['date']}\n")
        f.write(f"{'='*70}\n")

        # Display parameters
        f.write(f"{'Parameters':<20}\n")
        f.write(f"{'-'*70}\n")
        for key, value in params.items():
            f.write(f"{key:<20} : {value}\n")
        f.write(f"{'='*70}\n")

        # Display results
        f.write(f"{'Task':<50} {'Time (seconds)':>20}\n")
        f.write(f"{'-'*70}\n")
        f.write(f"{'Computing the solenoidal velocity field':<50} {results['solenoidal_time']:>20.4f}\n")
        f.write(f"{'Computing the incompressible pressure fluctuations':<50} {results['pressure_time']:>20.4f}\n")
        f.write(f"{'Computing the dilatational velocity field':<50} {results['dilatational_time']:>20.4f}\n")
        f.write(f"{'Computing the thermodynamic fields':<50} {results['thermodynamic_time']:>20.4f}\n")
        f.write(f"{'Plotting the spectrum':<50} {results['spectrum_time']:>20.4f}\n")
        f.write(f"{'Saving the data':<50} {results['save_time']:>20.4f}\n")
        f.write(f"{'-'*70}\n")
        f.write(f"{'Total TKE':<50} {results['TKE']:>20.4f}\n")
        f.write(f"{'Expected TKE':<50} {results['Ek']:>20.4f}\n")
        f.write(f"{'='*70}\n")

    # Print log to console
    with open(log_file, "r") as f:
        print(f.read())
